from typing import Dict, List, Optional

from pydantic import BaseModel, Field, validator

from moby_distribution.registry.utils import validate_media_type
from moby_distribution.spec.base import Descriptor, Platform


class FileSystemLayer(BaseModel):
    blobSum: str


class JWS(BaseModel):
    header: Dict
    protected: str
    signature: str


class Schema1History(BaseModel):
    """V1Compatibility is the raw V1 compatibility information.
    This will contain the JSON object describing the V1 of this image."""

    v1Compatibility: str


class ManifestSchema1(BaseModel):
    """image manifest for the Registry, Schema1.
    spec: https://github.com/distribution/distribution/blob/main/docs/spec/manifest-v2-1.md
    """

    schemaVersion: int = 1
    name: str
    tag: str
    architecture: str
    fsLayers: List[FileSystemLayer]
    history: List[Schema1History]
    signatures: Optional[List[JWS]]

    @staticmethod
    def content_type() -> str:
        return "application/vnd.docker.distribution.manifest.v1+prettyjws"

    @validator("schemaVersion")
    def validate_schema_version(cls, v):
        if v != 1:
            raise ValueError("schema version of ManifestSchema1 MUST be 1")
        return v


class PlatformSpec(Platform):
    """
    PlatformSpec specifies a platform where a particular image manifest is applicable.
    """

    features: Optional[List[str]]


class DockerManifestConfigDescriptor(Descriptor):
    @staticmethod
    def content_type() -> str:
        return "application/vnd.docker.container.image.v1+json"

    mediaType: str = "application/vnd.docker.container.image.v1+json"
    _validate_media_type = validator("mediaType", allow_reuse=True)(validate_media_type)


class DockerManifestLayerDescriptor(Descriptor):
    @staticmethod
    def content_types() -> List[str]:
        return [
            "application/vnd.docker.image.rootfs.diff.tar.gzip",
            "application/vnd.docker.image.rootfs.foreign.diff.tar.gzip",
        ]

    mediaType: str = "application/vnd.docker.image.rootfs.diff.tar.gzip"
    _validate_media_type = validator("mediaType", allow_reuse=True)(validate_media_type)


class ManifestSchema2(BaseModel):
    """image manifest for the Registry, Schema2.
    spec: https://github.com/distribution/distribution/blob/main/docs/spec/manifest-v2-2.md"""

    schemaVersion: int = 2
    mediaType: str = "application/vnd.docker.distribution.manifest.v2+json"
    config: DockerManifestConfigDescriptor
    layers: List[DockerManifestLayerDescriptor]

    @staticmethod
    def content_type() -> str:
        return "application/vnd.docker.distribution.manifest.v2+json"

    @validator("schemaVersion")
    def validate_schema_version(cls, v):
        if v != 2:
            raise ValueError("schema version of ManifestSchema2 MUST be 2")
        return v

    _validate_media_type = validator("mediaType", allow_reuse=True)(validate_media_type)


class OCIManifestConfigDescriptor(Descriptor):
    @staticmethod
    def content_type() -> str:
        return "application/vnd.oci.image.config.v1+json"

    _validate_media_type = validator("mediaType", allow_reuse=True)(validate_media_type)


class OCIManifestLayerDescriptor(Descriptor):
    @staticmethod
    def content_types() -> List[str]:
        return [
            "application/vnd.oci.image.layer.v1.tar",
            "application/vnd.oci.image.layer.v1.tar+gzip",
            "application/vnd.oci.image.layer.nondistributable.v1.tar",
            "application/vnd.oci.image.layer.nondistributable.v1.tar+gzip",
        ]

    _validate_media_type = validator("mediaType", allow_reuse=True)(validate_media_type)


class OCIManifestSchema1(BaseModel):
    """image manifest for the OCI Image

    spec: https://github.com/opencontainers/image-spec/blob/main/manifest.md
    """

    schemaVersion: int
    mediaType: str = "application/vnd.oci.image.manifest.v1+json"
    config: OCIManifestConfigDescriptor
    layers: List[OCIManifestLayerDescriptor]
    annotations: Dict[str, str] = Field(default_factory=dict)

    @staticmethod
    def content_type() -> str:
        return "application/vnd.oci.image.manifest.v1+json"

    @validator("schemaVersion")
    def validate_schema_version(cls, v):
        if v != 2:
            raise ValueError("schema version of OCIManifestSchema1 MUST be 2")
        return v

    _validate_media_type = validator("mediaType", allow_reuse=True)(validate_media_type)


class ManifestDescriptor(Descriptor):
    """ManifestDescriptor references a platform-specific manifest."""

    platform: Optional[PlatformSpec]



# -*- coding: utf-8 -*-
import base64
import logging
from typing import Any, Dict, Optional

import requests
from www_authenticate import parse

from moby_distribution.registry.exceptions import AuthFailed
from moby_distribution.spec.auth import TokenResponse

AUTH_TIMEOUT = 60 * 3
logger = logging.getLogger(__name__)


class AuthorizationProvider:
    def provide(self) -> str:
        """Provide the 'Authorization' used in HTTP Headers

        Usage: AuthorizationProvider().provider()
        """
        raise NotImplementedError


class BaseAuthentication:
    """Base Authentication Protocol"""

    def __init__(self, www_authenticate: str):
        self._raw_www_authenticate = www_authenticate
        self._www_authenticate = None

    @property
    def www_authenticate(self):
        if self._www_authenticate is None:
            self._www_authenticate = parse(self._raw_www_authenticate)
        return self._www_authenticate

    def authenticate(self, username: Optional[str] = None, password: Optional[str] = None) -> AuthorizationProvider:
        raise NotImplementedError

    @property
    def raw_www_authenticate(self) -> str:
        return self._raw_www_authenticate


class BasicAuthAuthorizationProvider(AuthorizationProvider):
    """BasicAuthAuthorizationProvider provide the `HTTP Basic Authentication`"""

    def __init__(self, username: str, password: str):
        self.username = username
        self.password = password

    def provide(self) -> str:
        auth = base64.b64encode(f"{self.username}:{self.password}".encode()).decode()
        return f"Basic {auth}"


class TokenAuthorizationProvider(AuthorizationProvider):
    """TokenAuthorizationProvider provide the `Bearer Token` Authentication"""

    def __init__(self, token_response: TokenResponse, token_type: str = "Bearer"):
        self.token_response = token_response
        self.token_type = token_type

    def provide(self) -> str:
        if self.token_response.token:
            return f"{self.token_type} {self.token_response.token}"
        elif self.token_response.access_token:
            return f"{self.token_type} {self.token_response.access_token}"
        raise ValueError("Missing Token")


class HTTPBasicAuthentication(BaseAuthentication):
    """`HTTP Basic Authentication` Authenticator"""

    def authenticate(self, username: Optional[str] = None, password: Optional[str] = None) -> AuthorizationProvider:
        if username is None or password is None:
            raise AuthFailed(
                message="请提供用户名和密码",
                status_code=400,
                response=None,
            )
        return BasicAuthAuthorizationProvider(username, password)


class DockerRegistryTokenAuthentication(BaseAuthentication):
    """Docker Registry v2 authentication via central service

    spec: https://github.com/distribution/distribution/blob/main/docs/spec/auth/token.md
    """

    REQUIRE_KEYS = ["realm", "service"]
    backend: str
    service: str
    scope: str

    def __init__(self, www_authenticate: str, offline_token: bool = True):
        super().__init__(www_authenticate)
        self.offline_token = offline_token

        assert "bearer" in self.www_authenticate
        self.bearer = self.www_authenticate["bearer"]

        for key in self.REQUIRE_KEYS:
            assert key in self.bearer

        self.backend = self.bearer["realm"]
        self.service = self.bearer["service"]
        self.scope = self.bearer.get("scope", None)

    def authenticate(self, username: Optional[str] = None, password: Optional[str] = None) -> AuthorizationProvider:
        """Authenticate to the registry.
        If no username and password provided, will authenticate as the anonymous user.

        :param username: User name to authenticate as.
        :param password: User's password.
        :return:
        """
        params: Dict[str, Any] = {
            "service": self.service,
            "scope": self.scope,
            "client_id": username or "anonymous",
            "offline_token": self.offline_token,
        }
        headers = {}
        if username and password:
            auth = base64.b64encode(f"{username}:{password}".encode()).decode()
            headers["Authorization"] = f"Basic {auth}"
        elif any([username, password]) and not all([username, password]):
            logger.warning("请同时提供 username 和 password!")

        logger.info("sending authentication request to authorization service<%s>", self.backend)
        resp = requests.get(self.backend, headers=headers, params=params, timeout=AUTH_TIMEOUT)
        if resp.status_code != 200:
            raise AuthFailed(
                message="用户凭证校验失败, 请检查用户信息和操作权限",
                status_code=resp.status_code,
                response=resp,
            )
        return TokenAuthorizationProvider(TokenResponse(**resp.json()))


class UniversalAuthentication(BaseAuthentication):
    """An Auto auth backend, which will auto auth by `scheme` provided by www_authenticate"""

    def authenticate(self, username: Optional[str] = None, password: Optional[str] = None) -> AuthorizationProvider:
        if "basic" in self.www_authenticate:
            return HTTPBasicAuthentication(self.raw_www_authenticate).authenticate(username, password)
        elif "bearer" in self.www_authenticate:
            return DockerRegistryTokenAuthentication(self.raw_www_authenticate).authenticate(username, password)
        raise NotImplementedError("未支持的认证方式")



import datetime
from textwrap import dedent
from typing import Dict, List, Optional

from pydantic import BaseModel, Field


def default_created():
    dt = datetime.datetime.utcnow()
    return dt.replace(microsecond=0, tzinfo=datetime.timezone.utc)


class HealthConfig(BaseModel):
    Test: List[str] = Field(
        ...,
        description=dedent(
            """
        Test is the test to perform to check that the container is healthy.
        An empty slice means to inherit the default.
        The options are:
        {} : inherit healthcheck
        {"NONE"} : disable healthcheck
        {"CMD", args...} : exec arguments directly
        {"CMD-SHELL", command} : run command with system's default shell
    """
        ),
    )
    Interval: int = Field(..., description="the time to wait between checks.")
    Timeout: int = Field(..., description="the time to wait before considering the check to have hung.")
    StartPeriod: int = Field(
        ..., description="the period for the container to initialize before the retries starts to count down."
    )
    Retries: int = Field(
        ..., description="the number of consecutive failures needed to consider a container as unhealthy."
    )


class ContainerConfig(BaseModel):
    """ContainerConfig contains the configuration data when running a container using the image
    spec: https://github.com/opencontainers/image-spec/blob/main/config.md
    """

    User: str = Field(
        "", description="User that will run the command(s) inside the container, also support user:group"
    )
    Memory: Optional[int] = Field(default=None, description="Memory limit (in bytes)")
    MemorySwap: Optional[int] = Field(
        default=None, description=" Total memory usage (memory + swap); set `-1` to enable unlimited swap"
    )
    CpuShares: Optional[int] = Field(default=None, description="CPU shares (relative weight vs. other containers)")
    ExposedPorts: Optional[Dict[str, Dict]] = Field(None, description="List of exposed ports")
    Env: Optional[List[str]] = Field(default=None, description="List of environment variable to set in the container")
    Entrypoint: Optional[List[str]] = Field(
        None, description="A list of arguments to use as the command to execute when the container starts. "
    )
    Cmd: Optional[List[str]] = Field(default=None, description="Default arguments to the entrypoint of the container.")
    Volumes: Optional[Dict] = Field(default=None, description="List of volumes (mounts) used for the container.")
    WorkingDir: Optional[str] = Field(
        default=None, description="Current directory (PWD) in the command will be launched"
    )
    Labels: Optional[Dict[str, str]] = Field(default=None, description="List of labels set to this container")
    StopSignal: Optional[str] = Field(
        default=None, description="The signal that will be sent to the container to exit."
    )
    Healthcheck: Optional[HealthConfig] = Field(
        default=None, description="Healthcheck describes how to check the container is healthy"
    )


class RootFS(BaseModel):
    diff_ids: List[str] = Field(..., description="diff id is the digest of uncompressed tarball")
    type: str = "layers"


class History(BaseModel):
    created: Optional[datetime.datetime] = Field(default_factory=default_created)
    author: Optional[str]
    created_by: Optional[str]
    comment: Optional[str]
    empty_layer: Optional[bool] = False


class ImageJSON(BaseModel):
    created: datetime.datetime
    author: str = "anonymous"
    architecture: str
    os: str
    variant: Optional[str] = Field(default=None, description="The variant of the specified CPU architecture.")
    config: ContainerConfig
    rootfs: RootFS
    history: List[History]



from typing import Optional, Union

import libtrust

from moby_distribution.registry.client import DockerRegistryV2Client, URLBuilder, default_client
from moby_distribution.registry.exceptions import ResourceNotFound, UnSupportMediaType
from moby_distribution.registry.resources import RepositoryResource
from moby_distribution.registry.utils import get_private_key
from moby_distribution.spec.manifest import ManifestDescriptor, ManifestSchema1, ManifestSchema2, OCIManifestSchema1


class ManifestRef(RepositoryResource):
    TYPES = {
        ManifestSchema1.content_type(): ManifestSchema1,
        ManifestSchema2.content_type(): ManifestSchema2,
        OCIManifestSchema1.content_type(): OCIManifestSchema1,
    }

    def __init__(
        self,
        repo: str,
        reference: str = "latest",
        client: DockerRegistryV2Client = default_client,
    ):
        super().__init__(repo, client)
        self.reference = reference

    def get(self, media_type: str = ManifestSchema2.content_type()):
        """retrieve image manifest as the provided media_type"""
        if media_type not in self.TYPES:
            raise UnSupportMediaType(media_type)

        type_ = self.TYPES[media_type]
        url = URLBuilder.build_manifests_url(self.client.api_base_url, self.repo, self.reference)
        headers = {"Accept": media_type}
        data = self.client.get(url=url, headers=headers).json()
        return type_(**data)

    def get_metadata(self, media_type: str = ManifestSchema2.content_type()) -> Optional[ManifestDescriptor]:
        """return ManifestDescriptor if the manifest exists."""
        if media_type not in self.TYPES:
            raise UnSupportMediaType(media_type)

        headers = {"Accept": media_type}
        url = URLBuilder.build_manifests_url(self.client.api_base_url, self.repo, self.reference)
        try:
            resp = self.client.head(url=url, headers=headers)
        except ResourceNotFound:
            return None

        media_type = resp.headers.get("Content-Type")
        digest = resp.headers.get("Docker-Content-Digest")
        size = resp.headers.get("Content-Length")

        return ManifestDescriptor(mediaType=media_type, digest=digest, size=size)

    def delete(self, raise_not_found: bool = True) -> bool:
        """Removes the manifest specified by the provided reference.

        Deleting a manifest that doesn't exist will raise ResourceNotFound if `raise_not_found` is true.
        Note that a manifest can only be deleted by digest,
        Therefore, delete a manifest will delete all tag associate with this manifest.
        """
        descriptor = self.get_metadata()
        if not descriptor:
            if raise_not_found:
                raise ResourceNotFound
            return False

        url = URLBuilder.build_manifests_url(self.client.api_base_url, self.repo, descriptor.digest)
        try:
            resp = self.client.delete(url=url)
        except ResourceNotFound:
            if raise_not_found:
                raise
            return False

        return resp.ok

    def put(self, manifest: Union[ManifestSchema1, ManifestSchema2, OCIManifestSchema1]) -> bool:
        """creates or updates the given manifest."""
        if isinstance(manifest, ManifestSchema1):
            resp = self._put_legacy_manifest(manifest)
        else:
            resp = self._put_new_manifest(manifest)

        return resp.ok

    def _put_legacy_manifest(self, manifest: ManifestSchema1):
        """put the docker schema 1 manifest with signed signature to the repository"""
        url = URLBuilder.build_manifests_url(self.client.api_base_url, self.repo, self.reference)
        private_key = get_private_key()
        data = manifest.json(
            include={
                "name",
                "tag",
                "architecture",
                "fsLayers",
                "history",
                "schemaVersion",
            }
        )
        js = libtrust.JSONSignature.new(data)
        js.sign(private_key)

        headers = {"Content-Type": manifest.content_type()}
        data = js.to_pretty_signature("signatures")
        return self.client.put(url=url, data=data, headers=headers)

    def _put_new_manifest(self, manifest: Union[ManifestSchema2, OCIManifestSchema1]):
        """put the docker schema 2 manifest or OCI manifest to the repository"""
        url = URLBuilder.build_manifests_url(self.client.api_base_url, self.repo, self.reference)
        headers = {"Content-Type": manifest.content_type()}
        data = manifest.json()
        return self.client.put(url=url, data=data, headers=headers)



from moby_distribution.registry.client import DockerRegistryV2Client, default_client, set_default_client
from moby_distribution.registry.resources.blobs import Blob
from moby_distribution.registry.resources.image import ImageRef, LayerRef
from moby_distribution.registry.resources.manifests import ManifestRef
from moby_distribution.registry.resources.tags import Tags
from moby_distribution.spec.endpoint import OFFICIAL_ENDPOINT, APIEndpoint
from moby_distribution.spec.image_json import ImageJSON
from moby_distribution.spec.manifest import ManifestSchema1, ManifestSchema2, OCIManifestSchema1

__version__ = "0.5.6"
__ALL__ = [
    "DockerRegistryV2Client",
    "Blob",
    "ManifestRef",
    "Tags",
    "APIEndpoint",
    "ManifestSchema1",
    "ManifestSchema2",
    "OCIManifestSchema1",
    "ImageJSON",
    "ImageRef",
    "LayerRef",
    "default_client",
    "set_default_client",
]



import logging
from functools import partial
from typing import Optional, Type, cast

import curlify
import requests

from moby_distribution.registry import exceptions
from moby_distribution.registry.auth import AuthorizationProvider, BaseAuthentication, UniversalAuthentication
from moby_distribution.registry.utils import LazyProxy
from moby_distribution.spec.endpoint import OFFICIAL_ENDPOINT, APIEndpoint

logger = logging.getLogger(__name__)


class DockerRegistryV2Client:
    """A Client implement APIs of Docker Registry HTTP API V2 and OCI Distribution Spec API

    spec: https://github.com/distribution/distribution/blob/main/docs/spec/api.md
    reference: https://github.com/distribution/distribution/tree/main/registry/client
    """

    @classmethod
    def from_api_endpoint(
        cls,
        api_endpoint: APIEndpoint = OFFICIAL_ENDPOINT,
        username: Optional[str] = None,
        password: Optional[str] = None,
        authenticator_class: Type[BaseAuthentication] = UniversalAuthentication,
    ):
        https_scheme = "https://"
        http_scheme = "http://"
        enable_https, certificate_valid = api_endpoint.is_secure_repository()
        if enable_https:
            client = cls(
                api_base_url=f"{https_scheme}{api_endpoint.api_base_url}",
                username=username,
                password=password,
                verify_certificate=certificate_valid,
                authenticator_class=authenticator_class,
                default_timeout=api_endpoint.default_timeout,
            )
            if certificate_valid or client.ping():
                return client
        return cls(
            api_base_url=f"{http_scheme}{api_endpoint.api_base_url}",
            username=username,
            password=password,
            verify_certificate=False,
            authenticator_class=authenticator_class,
            default_timeout=api_endpoint.default_timeout,
        )

    def __init__(
        self,
        api_base_url: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        verify_certificate: bool = True,
        authenticator_class: Type[BaseAuthentication] = UniversalAuthentication,
        default_timeout: float = 60 * 10,
    ):
        if api_base_url.endswith("/"):
            api_base_url = api_base_url.rstrip("/")
        self.api_base_url = api_base_url
        self.session = requests.session()
        self.session.verify = verify_certificate
        self.default_timeout = default_timeout

        self.username = username
        self.password = password
        self.authenticator_class = authenticator_class
        self._authed: Optional[AuthorizationProvider] = None

    def ping(self) -> bool:
        """API Version Check."""
        url = URLBuilder.build_v2_url(self.api_base_url)
        try:
            self._request(self.session.get, url=url)
        except exceptions.RequestError:
            logger.debug("Can't not connect to server<%s>", url)
            return False
        return True

    @property
    def authorization(self) -> str:
        if self._authed is None:
            return ""
        return self._authed.provide()

    @property
    def get(self):
        return partial(self._request, self.session.get)

    @property
    def put(self):
        return partial(self._request, self.session.put)

    @property
    def patch(self):
        return partial(self._request, self.session.patch)

    @property
    def post(self):
        return partial(self._request, self.session.post)

    @property
    def delete(self):
        return partial(self._request, self.session.delete)

    @property
    def head(self):
        return partial(self._request, self.session.head)

    def _request(self, method, *, should_retry: bool = True, **kwargs):
        kwargs.setdefault("timeout", self.default_timeout)
        headers = kwargs.setdefault("headers", {})
        headers["Authorization"] = self.authorization
        try:
            resp = self._validate_response(method(**kwargs), auto_auth=should_retry)
        except exceptions.RetryAgain:
            return self._request(method, should_retry=False, **kwargs)
        return resp

    def _validate_response(self, resp: requests.Response, auto_auth: bool = True) -> requests.Response:
        url = resp.request.url
        try:
            curl = curlify.to_curl(resp.request)
        except Exception:
            curl = "<unknown>"

        if resp.status_code == 401:
            if auto_auth:
                www_authenticate = resp.headers["www-authenticate"]
                auth = self.authenticator_class(www_authenticate)
                self._authed = auth.authenticate(username=self.username, password=self.password)
                raise exceptions.RetryAgain

            logger.debug("Requesting %s, but PermissionDeny, Equivalent curl command: %s", url, curl)
            raise exceptions.PermissionDeny

        if resp.status_code == 403:
            if auto_auth and self._authed is None and self.ping():
                raise exceptions.RetryAgain

            logger.debug("Requesting %s, but PermissionDeny, Equivalent curl command: %s", url, curl)
            raise exceptions.PermissionDeny

        if resp.status_code == 404:
            logger.info("Requesting %s, but ResourceNotFound, Equivalent curl command: %s", url, curl)
            raise exceptions.ResourceNotFound

        if not resp.ok:
            logger.warning("Requesting %s, but Response Not OK, Equivalent curl command: %s", url, curl)
            raise exceptions.RequestErrorWithResponse(message=resp.text, status_code=resp.status_code, response=resp)
        return resp


class URLBuilder:
    @staticmethod
    def build_v2_url(endpoint: str) -> str:
        return f"{endpoint}/v2/"

    @staticmethod
    def build_blobs_url(endpoint: str, repo: str, digest: str) -> str:
        return f"{endpoint}/v2/{repo}/blobs/{digest}"

    @staticmethod
    def build_manifests_url(endpoint: str, repo: str, reference: str) -> str:
        return f"{endpoint}/v2/{repo}/manifests/{reference}"

    @staticmethod
    def build_upload_blobs_url(endpoint: str, repo: str) -> str:
        return f"{endpoint}/v2/{repo}/blobs/uploads/"

    @staticmethod
    def build_tags_url(endpoint: str, repo: str) -> str:
        return f"{endpoint}/v2/{repo}/tags/list"


default_client = cast(
    DockerRegistryV2Client, LazyProxy(lambda: DockerRegistryV2Client.from_api_endpoint(OFFICIAL_ENDPOINT))
)


def set_default_client(client: DockerRegistryV2Client):
    default_client.__dict__["_wrapped"] = client



import datetime
from typing import Optional

from pydantic import BaseModel, Field


class TokenResponse(BaseModel):
    """Token Response
    spec: https://github.com/distribution/distribution/blob/main/docs/spec/auth/token.md#token-response-fields

    """

    token: str = Field(
        ...,
        description="An opaque Bearer token that clients should "
        "supply to subsequent requests in the Authorization header.",
    )
    access_token: Optional[str] = Field(
        None,
        description="For compatibility with OAuth 2.0, we will also accept token under the name access_token",
    )
    issued_at: Optional[datetime.datetime] = Field(
        None,
        description="The RFC3339-serialized UTC standard time at which a given token was issued.",
    )
    expires_in: Optional[int] = Field(
        None,
        description="The duration in seconds since the token was issued that it will remain valid. ",
    )
    refresh_token: Optional[str] = Field(
        None,
        description="Token which can be used to get additional access tokens "
        "for the same subject with different scopes. ",
    )



from typing import Optional

import requests


class RequestError(Exception):
    """Base Exception for requests"""

    def __init__(self, message: str, status_code):
        self.message = message
        super().__init__(message, status_code)


class RequestErrorWithResponse(RequestError):
    """Request error with requests.Response"""

    def __init__(self, message: str, status_code, response: Optional[requests.Response] = None):
        super().__init__(message, status_code)
        self.response = response


class AuthFailed(RequestErrorWithResponse):
    """Auth Failed for registry"""


class RetryAgain(Exception):
    """Dummy Exception to mark retry"""


class PermissionDeny(Exception):
    """Permission deny for endpoints or resources"""


class ResourceNotFound(Exception):
    """Resources not found."""


class UnSupportMediaType(Exception):
    """raise when the media type is unsupported"""



from typing import Dict, List, Optional

from pydantic import BaseModel, Field


class Platform(BaseModel):
    """
    Platform describes the platform which the image in the manifest runs on.
    """

    architecture: str
    os: str
    os_version: str = Field(alias="os.version")
    os_features: Optional[List[str]] = Field(alias="os.features")
    variant: str


class Descriptor(BaseModel):
    """
    Descriptor describes targeted content. Used in conjunction with a blob
    store, a descriptor can be used to fetch, store and target any kind of
    blob. The struct also describes the wire protocol format. Fields should
    only be added but never changed.

    spec: https://github.com/distribution/distribution/blob/cc4627fc6e5f20cfe8534492b44331fa16ccf872/blobs.go#L61
    see also: https://github.com/opencontainers/image-spec/blob/main/descriptor.md
    """

    mediaType: str
    size: int
    digest: str
    urls: List[str] = Field(default_factory=list)
    annotations: Dict[str, str] = Field(default_factory=dict)
    platform: Optional[Platform]



import hashlib
import io
import shutil
from contextlib import contextmanager
from pathlib import Path
from typing import IO, Optional, Tuple, Union
from urllib.parse import urlparse

from moby_distribution.registry import exceptions
from moby_distribution.registry.client import DockerRegistryV2Client, URLBuilder, default_client
from moby_distribution.registry.resources import RepositoryResource
from moby_distribution.spec.base import Descriptor


class Blob(RepositoryResource):
    def __init__(
        self,
        repo: str,
        digest: Optional[str] = None,
        local_path: Optional[Union[Path, str]] = None,
        fileobj: Optional[Union[IO, 'HashSignWrapper']] = None,
        client: DockerRegistryV2Client = default_client,
    ):
        super().__init__(repo, client)
        if isinstance(local_path, str):
            local_path = Path(local_path)

        self.digest = digest
        self.local_path = local_path
        self.fileobj = fileobj
        self._accessor = None

    @property
    def accessor(self):
        if self._accessor is None:
            self._accessor = Accessor(local_path=self.local_path, fileobj=self.fileobj)
        return self._accessor

    def stat(self, digest: Optional[str] = None) -> Descriptor:
        """Obtain resource information without receiving all data."""
        digest = digest or self.digest
        if digest is None:
            raise RuntimeError("unknown digest")

        url = URLBuilder.build_blobs_url(self.client.api_base_url, repo=self.repo, digest=digest)
        resp = self.client.head(url=url)
        headers = resp.headers
        return Descriptor(
            # Content-Type: application/octet-stream
            mediaType=headers["Content-Type"],
            # The Content-Length in headers is the `Descriptor Message` size, but not the `Blob` itself.
            size=headers.get("Content-Length", 0),
            digest=headers.get("Docker-Content-Digest", digest),
            urls=[headers.get("Location", url)],
        )

    def download(self, digest: Optional[str] = None):
        """download the blob from registry to `local_path` or `fileobj`"""
        digest = digest or self.digest
        if digest is None:
            raise RuntimeError("unknown digest")

        url = URLBuilder.build_blobs_url(self.client.api_base_url, repo=self.repo, digest=digest)
        resp = self.client.get(url=url, stream=True)
        with self.accessor.open(mode="wb") as fh:
            for chunk in resp.iter_content(chunk_size=1024):
                fh.write(chunk)

    def upload(self) -> Descriptor:
        """upload the blob from `local_path` or `fileobj` to the registry by streaming"""
        uuid, location = self._initiate_blob_upload()
        blob = BlobWriter(uuid, location, client=self.client)
        with self.accessor.open(mode="rb") as fh:
            signer = HashSignWrapper(fh=blob)
            shutil.copyfileobj(fsrc=fh, fdst=signer, length=1024 * 1024 * 64)

        digest = signer.digest()
        blob.commit(digest)
        self.digest = digest
        return self.stat()

    def upload_at_one_time(self) -> Descriptor:
        """upload the monolithic blob from `local_path` or `fileobj` to the registry at one time."""
        data = self.accessor.read_bytes()
        digest = f"sha256:{hashlib.sha256(data).hexdigest()}"

        headers = {"content_type": "application/octect-stream"}
        params = {"digest": digest}

        uuid, location = self._initiate_blob_upload()
        resp = self.client.put(url=location, headers=headers, params=params, data=data)

        if resp.status_code != 201:
            raise exceptions.RequestErrorWithResponse("failed to upload", status_code=resp.status_code, response=resp)
        self.digest = digest
        return self.stat()

    def _initiate_blob_upload(self) -> Tuple[str, str]:
        """Initiate a resumable blob upload.
        If successful, an uuid and upload location will be provided to complete the upload."""
        url = URLBuilder.build_upload_blobs_url(self.client.api_base_url, self.repo)
        resp = self.client.post(url=url)
        if resp.status_code != 202:
            raise exceptions.RequestError("Unexpected status code.", status_code=resp.status_code)

        uuid = resp.headers.get("docker-upload-uuid")
        location = resp.headers["location"]

        if uuid is None:
            uuid = location.split("/")[-1]

        if uuid == "":
            raise exceptions.RequestErrorWithResponse(
                "cannot retrieve docker upload UUID",
                status_code=resp.status_code,
                response=resp,
            )

        # Optionally, the location MAY be absolute (containing the protocol and/or hostname),
        # or it MAY be relative (containing just the URL path). For more information, see RFC 7231.
        if urlparse(location).netloc == "":
            location = f"{self.client.api_base_url}/{location.lstrip('/')}"
        return uuid, location

    def mount_from(self, from_repo: str) -> Descriptor:
        """Mount the blob from the given repo, if the client has read access to."""
        if self.digest is None:
            raise RuntimeError("unknown digest")

        url = URLBuilder.build_upload_blobs_url(self.client.api_base_url, self.repo)
        resp = self.client.post(url=url, params={"from": from_repo, "mount": self.digest})

        # If a registry does not support cross-repository mounting or is unable to mount the requested blob,
        # it SHOULD return a 202. At this time, we should upload the Blob to the registry.
        if resp.status_code == 202:
            return self._download_then_upload(from_repo=from_repo)

        # If the blob is successfully mounted, the client will receive a `201` Created response
        if resp.status_code != 201:
            raise exceptions.RequestErrorWithResponse(
                f"failed to mount blob({self.digest}) from `{from_repo}`", status_code=resp.status_code, response=resp
            )
        return self.stat()

    def delete(self, digest: Optional[str] = None):
        """Delete the blob identified by repo and digest"""
        digest = digest or self.digest
        if digest is None:
            raise RuntimeError("unknown digest")

        url = URLBuilder.build_blobs_url(self.client.api_base_url, repo=self.repo, digest=digest)
        resp = self.client.delete(url=url)
        if resp.status_code != 202:
            raise exceptions.RequestErrorWithResponse(
                f"failed to delete blob({self.digest}) from `{self.repo}`", status_code=resp.status_code, response=resp
            )
        return True

    def _download_then_upload(self, from_repo: str) -> Descriptor:
        """Fallback action for mount_from"""
        if self.fileobj is None and self.local_path is None:
            self.fileobj = fileobj = io.BytesIO()
            other = Blob(repo=from_repo, digest=self.digest, client=self.client, fileobj=fileobj)
            other.download()
            fileobj.seek(0)
        return self.upload()


class BlobWriter:
    def __init__(self, uuid: str, location: str, client: DockerRegistryV2Client):
        self.uuid = uuid
        self.location = location
        self.client = client
        self._committed = False
        self._offset = 0

    def write(self, buffer: Union[bytes, bytearray]) -> int:
        headers = {
            "content-range": f"{self._offset}-{self._offset + len(buffer) - 1}",
            "content-type": "application/octet-stream",
        }
        resp = self.client.patch(url=self.location, data=buffer, headers=headers)

        if resp.status_code != 202:
            raise exceptions.RequestErrorWithResponse(
                "fail to upload a chunk of blobs",
                status_code=resp.status_code,
                response=resp,
            )

        start_s, end_s = resp.headers["range"].split("-", 1)
        start, end = int(start_s), int(end_s)
        size = end - start + 1 - self._offset

        uuid = resp.headers.get("docker-upload-uuid")
        location = resp.headers["location"]

        if uuid is None:
            uuid = location.split("/")[-1]

        if uuid == "":
            raise exceptions.RequestErrorWithResponse(
                "cannot retrieve docker upload UUID",
                status_code=resp.status_code,
                response=resp,
            )

        self.uuid = uuid
        self.location = location
        self._offset += size
        return size

    def commit(self, digest: str) -> bool:
        params = {"digest": digest}
        resp = self.client.put(url=self.location, params=params)
        if resp.status_code != 201:
            raise exceptions.RequestErrorWithResponse(
                "can't commit an upload process",
                status_code=resp.status_code,
                response=resp,
            )
        self._committed = True
        return True

    def tell(self) -> int:
        return self._offset


class Accessor:
    def __init__(self, local_path: Optional[Path] = None, fileobj: Optional[IO] = None):
        if not local_path and not fileobj:
            raise ValueError("Nothing to operate")
        if local_path and fileobj:
            raise ValueError("Cannot be provided `local_path` and `fileobj` at the same time")
        self.local_path = local_path
        self.fileobj = fileobj

    @contextmanager
    def open(self, *args, **kwargs):
        if self.fileobj:
            yield self.fileobj
        else:
            with self.local_path.open(*args, **kwargs) as fh:
                yield fh

    def read_bytes(self):
        if self.fileobj:
            self.fileobj.seek(0)
            return self.fileobj.read()
        return self.local_path.read_bytes()


class CounterIO:
    def __init__(self):
        self.size = 0

    def write(self, chunk: bytes):
        self.size += len(chunk)
        return len(chunk)

    def tell(self) -> int:
        return self.size


class HashSignWrapper:
    """A Wrapper can sign the content of fh when copying it.

    Usage:
    >>> import shutil
    >>> src = open("somewhere", mode="rb")
    >>> dest = HashSignWrapper(open("somewhere", mode="wb"))
    >>> shutil.copyfileobj(src, dest)
    """

    def __init__(self, fh: Optional[Union[IO, CounterIO, BlobWriter]] = None, constructor=hashlib.sha256):
        self._raw_fh = fh or CounterIO()
        self.signer = constructor()

    def write(self, chunk: bytes):
        self.signer.update(chunk)
        return self._raw_fh.write(chunk)

    def tell(self) -> int:
        return self._raw_fh.tell()

    def digest(self) -> str:
        """return hexdigest with hash method name"""
        return f"{self.signer.name}:{self.signer.hexdigest()}"



import re
import socket
import ssl
from typing import Pattern, Tuple

from pydantic import BaseModel


class APIEndpoint(BaseModel):
    version: int = 2
    url: str
    official: bool = False
    default_timeout: float = 60 * 10

    def is_secure_repository(self) -> Tuple[bool, bool]:
        """Detect if the repository is secure

        :returns Tuple[bool, bool], the first one mean if the server support https?,
                                    the second one mean if the ssl certificate is valid?
        """
        match = url_regex().match(self.url)
        if not match:
            return False, False

        parts = match.groupdict()
        hostname = parts.get("domain") or parts.get("ipv4") or parts.get("ipv6")
        port = int(parts.get("port") or 443)
        try:
            context = ssl.create_default_context()
            connection = socket.create_connection((hostname, port), timeout=self.default_timeout)
            sock = context.wrap_socket(connection, server_hostname=hostname)
            sock.getpeercert()
        except ssl.SSLError as e:
            if e.reason == "CERTIFICATE_VERIFY_FAILED":
                return True, False
            elif e.reason == "WRONG_VERSION_NUMBER":
                return False, False
            return False, False
        except socket.timeout:
            return False, False
        except OSError:
            return False, False
        return True, True

    @property
    def api_base_url(self) -> str:
        match = url_regex().match(self.url)
        if not match:
            raise ValueError("Invalid Url")

        parts = match.groupdict()
        hostname = parts["domain"] or parts["ipv4"] or parts["ipv6"]

        port = parts["port"]
        if not port:
            port = "443" if self.is_secure_repository()[0] else "80"

        path = parts["path"] or ""
        return f"{hostname}:{port}{path}"


_url_regex_cache = None


def url_regex() -> Pattern[str]:
    global _url_regex_cache
    if _url_regex_cache is None:
        _url_regex_cache = re.compile(
            r"(?:(?P<scheme>[a-z][a-z0-9+\-.]+)://)?"  # scheme https://tools.ietf.org/html/rfc3986#appendix-A
            r"(?:(?P<user>[^\s:/]*)(?::(?P<password>[^\s/]*))?@)?"  # user info
            r"(?:"
            r"(?P<ipv4>(?:\d{1,3}\.){3}\d{1,3})|"  # ipv4
            r"(?P<ipv6>\[[A-F0-9]*:[A-F0-9:]+\])|"  # ipv6
            r"(?P<domain>[^\s/:?#]+)"  # domain, validation occurs later
            r")?"
            r"(?::(?P<port>\d+))?"  # port
            r"(?P<path>/[^\s?#]*)?"  # path
            r"(?:\?(?P<query>[^\s#]+))?"  # query
            r"(?:#(?P<fragment>\S+))?",  # fragment
            re.IGNORECASE,
        )
    return _url_regex_cache


OFFICIAL_ENDPOINT = APIEndpoint(url="registry.hub.docker.com", official=True)



import gzip
import hashlib
import io
import json
import logging
import shutil
import tarfile
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import List, Optional

from pydantic import BaseModel, Field

from moby_distribution.registry.client import DockerRegistryV2Client, default_client
from moby_distribution.registry.resources import RepositoryResource
from moby_distribution.registry.resources.blobs import Blob, HashSignWrapper
from moby_distribution.registry.resources.manifests import ManifestRef
from moby_distribution.registry.utils import generate_temp_dir, parse_image
from moby_distribution.spec.image_json import History, ImageJSON, default_created
from moby_distribution.spec.manifest import (
    DockerManifestConfigDescriptor,
    DockerManifestLayerDescriptor,
    ManifestSchema2,
)

logger = logging.getLogger(__name__)


class ImageManifest(BaseModel):
    config: str = Field(default="", alias="Config")
    RepoTags: List[str] = Field(default_factory=list)
    Layers: List[str] = Field(default_factory=list)


class LayerRef(BaseModel):
    repo: str = ""
    digest: str = ""
    size: int = -1
    exists: bool = False
    local_path: Optional[Path]


class ImageRef(RepositoryResource):
    """ImageRef is used to Manipulate Docker images"""

    def __init__(
        self,
        repo: str,
        reference: str,
        layers: List[LayerRef],
        initial_config: str,
        client: DockerRegistryV2Client = default_client,
    ):
        super().__init__(repo, client)
        self.reference = reference
        self.layers = layers
        self._initial_config = initial_config
        self._dirty = False
        # diff id is the digest of uncompressed tarball
        self._append_diff_ids: List[str] = []
        self._append_historys: List[History] = []

    @classmethod
    def from_image(
        cls,
        from_repo: str,
        from_reference: str,
        to_repo: Optional[str] = None,
        to_reference: Optional[str] = None,
        client: DockerRegistryV2Client = default_client,
    ):
        """Initial a `ImageRef` from `{from_repo}:{from_reference}` but will named it as `{to_repo, to_reference}`

        if no `to_repo` or `to_reference` given, use `from_repo` or `from_reference` as default.
        """
        if to_repo is None:
            to_repo = from_repo
        if to_reference is None:
            to_reference = from_reference
        manifest = ManifestRef(repo=from_repo, reference=from_reference, client=client).get(
            ManifestSchema2.content_type()
        )
        layers = [
            LayerRef(repo=from_repo, digest=layer.digest, size=layer.size, exists=True) for layer in manifest.layers
        ]

        fh = io.BytesIO()
        Blob(repo=from_repo, digest=manifest.config.digest, client=client, fileobj=fh).download()
        fh.seek(0)

        return cls(
            repo=to_repo, reference=to_reference, layers=layers, initial_config=fh.read().decode(), client=client
        )

    @classmethod
    def from_tarball(
        cls,
        workplace: Path,
        src: Path,
        to_repo: Optional[str] = None,
        to_reference: Optional[str] = None,
        client: DockerRegistryV2Client = default_client,
    ):
        """Initial a `ImageRef` from a tarball locate in local disk, but will named it as `{to_repo, to_reference}`

        :param Path workplace: the workplace to extract the tarball and store gzip compressed layers
        :param Path src: the path of tarball
        :param str to_repo: the name for the image
        :param str to_reference: the tag for the image

        if no `to_repo` or `to_reference` given, will use RepoTags in `manifest.json`
        """
        with tarfile.open(name=src) as tarball:
            tarball.extractall(workplace)

            manifest_list = json.loads((workplace / "manifest.json").read_text())
            if not isinstance(manifest_list, list) or len(manifest_list) == 0:
                raise ValueError("Invalid manifest.json")

            manifest = ImageManifest(**manifest_list[0])
            if not manifest.RepoTags and (not to_repo or not to_reference):
                raise ValueError("Invalid repo or reference")

            named = parse_image(manifest.RepoTags[0])
            if to_repo is None:
                to_repo = named.name

            if to_reference is None:
                to_reference = named.tag or "latest"

            layers = []
            for layer in manifest.Layers:
                gzipped_filepath = workplace / (layer + ".gz")
                with (workplace / layer).open(mode="rb") as fh, gzip.open(gzipped_filepath, mode="wb") as compressed:
                    shutil.copyfileobj(fh, compressed)

                # The gzipped file can only be obtained after the compressed object is closed
                # (because the gzip context information has not yet been written).
                gzipped_signer = HashSignWrapper()
                with gzipped_filepath.open(mode="rb") as fh:
                    shutil.copyfileobj(fh, gzipped_signer)

                layers.append(
                    LayerRef(
                        repo=to_repo,
                        digest=gzipped_signer.digest(),
                        size=gzipped_signer.tell(),
                        local_path=gzipped_filepath,
                    )
                )

            return cls(
                repo=to_repo,
                reference=to_reference,
                layers=layers,
                initial_config=(workplace / manifest.config).read_text(),
                client=client,
            )

    def save(self, dest: str):
        """save the image to dest, as Docker Image Specification v1.2 Format

        spec: https://github.com/moby/moby/blob/master/image/spec/v1.2.md
        """
        manifest = ImageManifest(RepoTags=[f"{self.repo}:{self.reference}"])
        with generate_temp_dir() as workplace:
            # Step 1. save image json
            image_json_digest = hashlib.sha256(self.image_json_str.encode()).hexdigest()

            manifest.config = f"{image_json_digest}.json"
            (workplace / manifest.config).write_text(self.image_json_str)

            # Step 2. download layers
            for layer in self.layers:
                manifest.Layers.append(self._save_layer(workplace, layer=layer))

            # Step 3. save manifest
            (workplace / "manifest.json").write_text(f"[{manifest.json(by_alias=True)}]")

            # Step 4. save as tar
            with tarfile.open(mode="w", name=dest) as tarball:
                for f in workplace.iterdir():
                    tarball.add(name=str(f.absolute()), arcname=str(f.relative_to(workplace)))
        return dest

    def push(self, media_type: str = ManifestSchema2.content_type(), *, max_worker: int = 5):
        """push the image to the registry."""
        if media_type == ManifestSchema2.content_type():
            return self.push_v2(max_worker=max_worker)
        raise NotImplementedError("only support push images with Manifest Schema2.")

    def push_v2(self, *, max_worker: int = 5) -> ManifestSchema2:
        """push the image to the registry, with Manifest Schema2."""
        layer_descriptors_futures = []
        layer_descriptors = []
        # Step 1: upload all layers
        with ThreadPoolExecutor(max_workers=max_worker) as thread_pool:
            for layer in self.layers:
                layer_descriptors_futures.append(thread_pool.submit(self._upload_layer, layer))
        for future in layer_descriptors_futures:
            layer_descriptors.append(future.result())

        # Step 2: upload the image json
        config_descriptor = self._upload_config(self.image_json_str)

        # Step 3.: upload the manifest
        manifest = ManifestSchema2(config=config_descriptor, layers=layer_descriptors)
        ref = ManifestRef(repo=self.repo, reference=self.reference, client=self.client)
        ref.put(manifest)
        if self._dirty:
            return ref.get(media_type=ManifestSchema2.content_type())
        return manifest

    def add_layer(self, layer: LayerRef, history: Optional[History] = None) -> DockerManifestLayerDescriptor:
        """Add a layer to this image.

        Step:
          1. calculate the sha256 sum for the gzipped_tarball, as digest
          2. calculate the sha256 sum for the uncompressed_tarball, as diff_id
        """
        if not layer.exists and not layer.local_path:
            raise ValueError("Unknown layer")

        uncompressed_tarball_signer = HashSignWrapper()
        # Add local layer
        if layer.local_path:
            # Step 1: calculate the sha256 sum for the gzipped_tarball
            gzipped_signer = HashSignWrapper()
            with layer.local_path.open(mode="rb") as gzipped:
                shutil.copyfileobj(gzipped, gzipped_signer)
                size = gzipped_signer.tell()

            # Step 2: calculate the sha256 sum for the uncompressed_tarball
            with gzip.open(filename=layer.local_path) as uncompressed:
                shutil.copyfileobj(uncompressed, uncompressed_tarball_signer)

            if layer.digest and layer.digest != gzipped_signer.digest():
                raise ValueError(
                    "Wrong digest, layer.digest<'%s'> != signer.digest<'%s'>",
                    layer.digest,
                    gzipped_signer.digest(),
                )

            layer.digest = gzipped_signer.digest()
            layer.repo = self.repo
            layer.size = size

        # Add remote layer if the layer is exists in registry
        else:
            with generate_temp_dir() as temp_dir:
                # Step 1: calculate the sha256 sum for the gzipped_tarball
                with (temp_dir / "blob").open(mode="wb") as fh:
                    gzipped_signer = HashSignWrapper(fh=fh)
                    Blob(repo=layer.repo, digest=layer.digest, fileobj=gzipped_signer, client=self.client).download()
                    size = gzipped_signer.tell()

                # Step 2: calculate the sha256 sum for the uncompressed_tarball
                with gzip.open(filename=(temp_dir / "blob")) as uncompressed:
                    shutil.copyfileobj(uncompressed, uncompressed_tarball_signer)

            if layer.size != size:
                raise ValueError("Wrong Size, layer.size<'%d'> != signer.size<'%d'>", layer.size, size)
            if layer.digest != gzipped_signer.digest():
                raise ValueError(
                    "Wrong digest, layer.digest<'%s'> != signer.digest<'%s'>",
                    layer.digest,
                    gzipped_signer.digest(),
                )

        self._dirty = True
        self._append_diff_ids.append(uncompressed_tarball_signer.digest())
        self._append_historys.append(
            history
            or History(
                comment="add by moby-distribution",
                created_by="add by moby-distribution",
                created=default_created(),
                empty_layer=False,
            )
        )
        self.layers.append(layer)

        return DockerManifestLayerDescriptor(
            digest=gzipped_signer.digest(),
            size=size,
        )

    @property
    def image_json(self) -> ImageJSON:
        base = ImageJSON(**json.loads(self._initial_config))
        if not self._dirty:
            return base
        base.rootfs.diff_ids.extend(self._append_diff_ids)
        base.history.extend(self._append_historys)
        return base

    @property
    def image_json_str(self) -> str:
        if not self._dirty:
            return self._initial_config
        image_json = self.image_json
        return image_json.json(exclude_unset=True, exclude_defaults=True, separators=(",", ":"))

    def _save_layer(self, workplace: Path, layer: LayerRef) -> str:
        """Download the gzipped layer, and uncompress as the raw tarball.

        if layer is exists in local disk(the local_path is not None), will skip download.

        :raise RequestErrorWithResponse: raise if an error occur.
        """
        gzip_path = layer.local_path or (workplace / "layers.tar.gz")
        temp_tarball_path = workplace / "layer.tar"

        if layer.local_path is None:
            Blob(repo=layer.repo, digest=layer.digest, local_path=gzip_path, client=self.client).download()

        with gzip.open(filename=gzip_path) as uncompressed, temp_tarball_path.open(mode="wb") as fh:
            signer = HashSignWrapper(fh)
            shutil.copyfileobj(uncompressed, signer)

        tarball_path = f"{signer.digest()}/layer.tar"
        (workplace / tarball_path).parent.mkdir(exist_ok=True, parents=True)

        if layer.local_path is None:
            gzip_path.unlink()

        shutil.move(str(temp_tarball_path.absolute()), str((workplace / tarball_path).absolute()))
        return tarball_path

    def _upload_layer(self, layer: LayerRef) -> DockerManifestLayerDescriptor:
        """Upload the layer to the registry
        this func will mount the existed layers from other repo or upload the local layers to the repo.

        :raise RequestErrorWithResponse: raise if an error occur.
        """

        if layer.exists and layer.repo != self.repo:
            descriptor = Blob(repo=self.repo, digest=layer.digest, client=self.client).mount_from(from_repo=layer.repo)
        elif not layer.exists:
            blob = Blob(repo=self.repo, local_path=layer.local_path, client=self.client)
            descriptor = blob.upload()
        else:
            descriptor = Blob(repo=self.repo, client=self.client).stat(layer.digest)

        return DockerManifestLayerDescriptor(
            size=layer.size,
            digest=descriptor.digest,
            urls=descriptor.urls,
        )

    def _upload_config(self, image_json_str: str) -> DockerManifestConfigDescriptor:
        """Upload the Image JSON to the registry

        :raise RequestErrorWithResponse: raise if an error occur.
        """
        descriptor = Blob(repo=self.repo, fileobj=io.BytesIO(image_json_str.encode()), client=self.client).upload()
        return DockerManifestConfigDescriptor(
            size=len(image_json_str),
            digest=descriptor.digest,
            urls=descriptor.urls,
        )



from moby_distribution.registry.client import DockerRegistryV2Client, default_client


class RepositoryResource:
    def __init__(
        self,
        repo: str,
        client: DockerRegistryV2Client = default_client,
    ):
        self.repo = repo
        if client is not None:
            self._client = client

    @property
    def client(self):
        if hasattr(self, "_client"):
            return self._client
        raise RuntimeError("Resource must bind Client")

    @client.setter
    def client(self, v: DockerRegistryV2Client):
        self._client = v



import logging
import os
import shutil
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Callable, ContextManager, Iterator, NamedTuple, Optional, Union
from urllib.parse import urlparse

import libtrust
from libtrust.keys import ec_key, rs_key

logger = logging.getLogger(__name__)


def get_private_key() -> Union[libtrust.ECPrivateKey, libtrust.RSAPrivateKey]:
    key = os.getenv("MOBY_DISTRIBUTION_PRIVATE_KEY")
    password = os.getenv("MOBY_DISTRIBUTION_PRIVATE_KEY_PASSWORD")
    if key is not None:
        try:
            return ec_key.ECPrivateKey.from_pem(key, password)
        except Exception:
            try:
                return rs_key.RSAPrivateKey.from_pem(key, password)
            except Exception:
                logger.warning("Unknown private key.")
    return ec_key.generate_private_key()


def validate_media_type(cls, media_type: str) -> str:
    if hasattr(cls, "content_type") and cls.content_type() != media_type:
        raise ValueError("unknown media type '{}'".format(media_type))

    if hasattr(cls, "content_types") and media_type not in cls.content_types():
        raise ValueError("unknown media type '{}'".format(media_type))

    return media_type


def new_method_proxy(func):
    def inner(self, *args):
        if "_wrapped" not in self.__dict__:
            self.__dict__["_wrapped"] = self.__dict__["_factory"]()
        return func(self._wrapped, *args)

    return inner


class LazyProxy:
    def __init__(self, obj: Callable[..., Any]):
        self.__dict__["_factory"] = obj

    __getattr__ = new_method_proxy(getattr)
    __setattr__ = new_method_proxy(setattr)
    __dir__ = new_method_proxy(dir)

    def __repr__(self):
        return repr(self.__dict__["_wrapped"])


def __generate_temp_dir__(suffix=None) -> Iterator[Path]:
    path = None
    try:
        path = Path(tempfile.mkdtemp(suffix=suffix))
        logger.debug('Generating temp path: %s', path)
        yield path
    finally:
        if path and path.exists():
            shutil.rmtree(path)


generate_temp_dir: Callable[..., ContextManager[Path]] = contextmanager(__generate_temp_dir__)


class NamedImage(NamedTuple):
    domain: str
    name: str
    tag: Optional[str] = None


def parse_image(image: str, default_registry: Optional[str] = None) -> NamedImage:
    """parse the `repository:tag` as NamedImage

    Usage:
    >>> parse_image("python", default_registry="docker.io")
    NamedImage(domain='docker.io', name='library/python', tag=None)

    >>> parse_image("python:latest", default_registry="docker.io")
    NamedImage(domain='docker.io', name='library/python', tag="latest")

    >>> parse_image("localhost:5000/python", default_registry="docker.io")
    NamedImage(domain='localhost:5000', name='python', tag=None)

    >>> parse_image("localhost:5000/python:latest", default_registry="docker.io")
    NamedImage(domain='localhost:5000', name='python', tag='latest')

    >>> parse_image("docker.io:5000/python:latest", default_registry="not-docker.io")
    NamedImage(domain='docker.io:5000', name='python', tag='latest')
    """
    from moby_distribution.registry.client import default_client

    i = image.find("/")
    default_registry = default_registry or urlparse(default_client.api_base_url).netloc
    tag: Optional[str] = None

    # case for image in default registry
    if i == -1 or (("." not in image[:i] and ":" not in image[:i]) and image[:i] != "localhost"):
        domain = default_registry
        remainder = image
    else:
        domain = image[:i]
        remainder = image[i + 1 :]

    if ":" in remainder:
        name, tag = remainder.split(":", 1)
    else:
        name = remainder

    if domain == default_registry and "/" not in name:
        name = f"library/{name}"

    return NamedImage(domain=domain, name=name, tag=tag)



from typing import List, Optional

from moby_distribution.registry.client import URLBuilder
from moby_distribution.registry.resources import RepositoryResource
from moby_distribution.registry.resources.manifests import ManifestRef
from moby_distribution.spec.manifest import ManifestDescriptor


class Tags(RepositoryResource):
    def get(self, tag: str) -> Optional[ManifestDescriptor]:
        """retrieve the ManifestDescriptor identified by the tag."""
        return ManifestRef(self.repo, reference=tag, client=self.client).get_metadata()

    def untag(self, tag: str) -> bool:
        """Untag removes the provided tag association"""
        return ManifestRef(self.repo, reference=tag, client=self.client).delete()

    def list(self) -> List[str]:
        """return the list of tags in the repo"""
        url = URLBuilder.build_tags_url(self.client.api_base_url, self.repo)
        data = self.client.get(url=url).json()
        return data["tags"] or []
